
INCLUDE 'kind.f90'
INCLUDE 'matrixoperations_dp.f90'
INCLUDE 'IO.f90'


PROGRAM bayes_pspline_hetero
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!                                                                        !!! 
!!!                                                                        !!! 
!!!                                                                        !!! 
!!!   * SEE THE "README.TXT" FILE TO GET INSTRUCTIONS ON USING THIS        !!!
!!!     PROGRAM AND INSTALLING SUPPORT LIBRARY CALLED "RANDLIB90"          !!!
!!!                                                                        !!!
!!!                                                                        !!!
!!!                                                                        !!!
!!!   * VARIABLE NAMES USED INSIDE THIS CODE ARE VERY SIMILAR TO           !!!
!!!     CORRESPONDING NAMES IN THE PAPER.                                  !!!
!!!                                                                        !!!
!!!                                                                        !!!
!!!                                                                        !!!
!!!   * THE METROPOLIS STEPS 6 AND 7 NEED TO BE TUNED IN ORDER TO          !!!
!!!     GET THE ACCEPTANCE RATIO IN A REASONABLE RANGE. THE                !!!
!!!     CORRESPONDING VARIABLE IN THIS CODE IS CALLED "mhprop". THE        !!!
!!!     OUTPUT FILES "metro6" AND "metro7" WOULD CONTAIN INFORMATION       !!!
!!!     ABOUT ACCEPTANCE RATES FOR A PARTICULAR SETTING                    !!!
!!!                                                                        !!!
!!!                                                                        !!! 
!!!                                                                        !!! 
!!!   * THIS PROGRAM CALLS ROUTINES FROM THE MODULE "matrix"               !!!
!!!     INSIDE THE FILE "matrixoperations_dp.f90" SEVERAL TIMES.           !!!
!!!     THOSE ROUTINES HAVE BEEN DESIGNED SPECIFICALLY TO ACHIEVE          !!!
!!!     FASTER COMPUTATION                                                 !!!
!!!                                                                        !!! 
!!!                                                                        !!! 
!!!                                                                        !!! 
!!!   * DESCRIPTION OF SOME OF THE VARIABLE NAMES ARE AS FOLLOWS:          !!!
!!!                                                                        !!! 
!!!     X, Y         DATA; VECTORS OF LENGTH "n"                           !!!
!!!     n            LENGTH OF THE OBSERVED VECTOR "Y"                     !!! 
!!!     Ksup*        NUMBER OF KNOTS IN THE SPLINES                        !!!
!!!     theta*       COEFFICIENTS OF THE SPLINES                           !!!
!!!     Csub*        DESIGN MATRIX                                         !!!
!!!     metro*       ACCEPTANCE RATE IN STEP *                             !!!
!!!     sam_*        SAMPLES OF VARIABLE "*"                               !!!
!!!                                                                        !!! 
!!!     THE FOLLOWING HAVE BEEN IMPLEMENTED TO ENHANCE COMPUTATION         !!!
!!!     trans*       TRANSPOSE OF MATRIX "*"                               !!!
!!!     trans*by**   TRANSPOSE OF MATRIX "*" DIVIDED BY "**"               !!!
!!!     inv_*        INVERSE OF MATRIX "*"                                 !!!
!!!     CTC#by@      TRANSPOSE(C_#) * C_# /@                               !!!
!!!     reci_*       RECIPROCAL OF PARAMETER "*"                           !!!
!!!                                                                        !!! 
!!!                                                                        !!! 
!!!                                                                        !!! 
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  use user_set_generator
  use random_standard_normal_mod
  use random_normal_mod
  use random_gamma_mod
  use random_standard_uniform_mod
  use kind
  use matrix
  use IO
  IMPLICIT NONE

  REAL(kn), ALLOCATABLE, DIMENSION(:) :: thetasubX,thetasubepsilon, thetasubb, Y, knotsm, knotsb
  REAL(kn), ALLOCATABLE, DIMENSION(:) :: sigma2epsilon,mu,eta,sigma2b,si,X
  REAL(kn), ALLOCATABLE, DIMENSION(:,:) :: CsubX, Csubepsilon ! Csubepsilon is the same as CsubX
  REAL(kn), ALLOCATABLE, DIMENSION(:,:) :: transCsubX,transCsubepsilonbysigma2u,Csubb,transCsubbbysigma2v
  REAL(kn), ALLOCATABLE, DIMENSION(:,:) :: inv_MsubX,inv_Msubepsilon,CTCepsilonbysigma2u ! CTCepsilonbysigma2u stores (transpose(Csubepsilon)*Csubepsilon)/sigma2u
  REAL(kn), ALLOCATABLE, DIMENSION(:,:) :: inv_SIGMAsub0X,inv_SIGMAsub0epsilon,inv_SIGMAsub0b, inv_Msubb,CTCbbysigma2v ! CTCbbysigma2v stores (transpose(Csubb)*Csubb)/sigma2v 


  REAL(kn) :: sigma2u = 0.01, sigma2v = 0.01, asubc = 1.e-3, asubd = 1.e-3, bsubc = 1.e-3, bsubd = 1.e-3
  REAL(kn), PARAMETER :: pi = 22./7.  ! declared constant
  REAL(kn) :: sigma20beta = 1.e6, sigma20gamma = 1.e6, sigma20delta = 1.e6  
  INTEGER :: p =2, q =1, Ksupepsilon =40, Ksupb =4, n =1000, max_iter = 120000, burn_in = 20000, thin = 10 
  REAL(kn) :: reci_sigma2c, reci_sigma2d, mhprop, a, b, c, d, e, u, gengamvar1, gengamvar2
  INTEGER :: main_iter, iter1, iter2, save_iter = 0, sim_iter ! iteration variables
  INTEGER :: zz1, zz2, saved
  REAL :: TIME

  REAL(kn), ALLOCATABLE, DIMENSION(:,:) :: TEMP_21, TEMP_5, TEMP_6, TEMP_7 
  REAL(kn), ALLOCATABLE, DIMENSION(:) :: TEMP_2, TEMP_3, TEMP_1, TEMP_4
  CHARACTER(len = 50):: fname1, fname2, fname3, fname4, fname5


!!! Arrays for saving samples

  REAL(kn), ALLOCATABLE, DIMENSION(:,:) :: sam_thetasubX, sam_thetasubepsilon, sam_thetasubb ! saved samples of theta vectors
  REAL(kn), ALLOCATABLE, DIMENSION(:) :: sam_sigma2c, sam_sigma2d
  REAL(kn), ALLOCATABLE, DIMENSION(:,:) :: sam_sigma2epsilon, sam_sigma2b

  REAL(kn), ALLOCATABLE, DIMENSION(:) :: metro6, metro7


  WRITE(*,*)""
  WRITE(*,*) "Note: Mean and Variance Functions are modeled with splines of same degree, also they share the design matrix"
  WRITE(*,*) "Enter the appropriate values in the steps to follow"
  WRITE(*,*) ""

  WRITE(*,*) "Length of X, degree of mean function spline, degree of penalty function spline(n,p,q)(default:",n,",",p,",",q,")"
  READ(*,*) n, p, q

  WRITE(*,*) "Number of knots for the mean and penalty functions (K^m,K^b)(default:",Ksupepsilon,",",Ksupb,")"
  READ(*,*) Ksupepsilon,Ksupb

  WRITE(*,*) "Number of samples to be drawn, burn-in, and thinning parameter (default:", max_iter,",", burn_in,",", thin,")"
  READ(*,*) max_iter,burn_in,thin

  WRITE(*,*) ""
  WRITE(*,*) "Now enter the model parameters. Refer to the paper for names"

  WRITE(*,100) sigma2u,sigma2v
100 FORMAT ('\sigma^2_u, \sigma^2_v (default:',F5.3,',',F5.3,')')
  READ(*,*) sigma2u, sigma2v

  WRITE(*,110) asubc,asubd,bsubc,bsubd
110 FORMAT ('a_c, a_d, b_c, b_d (default:',F5.3,',',F5.3,',',F5.3,',',F5.3,')')
  READ(*,*) asubc,asubd,bsubc,bsubd

  WRITE(*,120) sigma20beta,sigma20gamma,sigma20delta
120 FORMAT ('\sigma^2_{\beta}, \sigma^2_{\gamma}, \sigma^2_{\delta} (default:', E9.2,',',E9.2,',',E9.2,')')
  READ(*,*) sigma20beta,sigma20gamma,sigma20delta

  saved = (max_iter - burn_in)/thin ! number of saved samples


  zz1 = Ksupepsilon + p + 1; zz2 = Ksupb + q + 1 



!!!!!!!!!!! Allocating Allocatable arrays

  ALLOCATE(thetasubX(zz1),thetasubepsilon(zz1),thetasubb(zz2),Y(n))
  ALLOCATE(sigma2epsilon(n),mu(n),eta(n),X(n),knotsm(Ksupepsilon),knotsb(Ksupb))
  ALLOCATE(sigma2b(Ksupepsilon),si(Ksupepsilon),CsubX(n,zz1),Csubepsilon(n,zz1))
  ALLOCATE(transCsubX(zz1,n),transCsubepsilonbysigma2u(zz1,n),Csubb(Ksupepsilon,zz2)) 
  ALLOCATE(transCsubbbysigma2v(zz2,Ksupepsilon))
  ALLOCATE(inv_MsubX(zz1,zz1),inv_Msubepsilon(zz1,zz1),CTCepsilonbysigma2u(zz1,zz1)) 
  ALLOCATE(inv_SIGMAsub0X(zz1,zz1),inv_SIGMAsub0epsilon(zz1,zz1),inv_SIGMAsub0b(zz2,zz2)) ! don't forget to initialize these to 0
  ALLOCATE(inv_Msubb(zz2,zz2),CTCbbysigma2v(zz2,zz2))


  inv_SIGMAsub0X = 0.; inv_SIGMAsub0epsilon = 0.; inv_SIGMAsub0b = 0.



  ALLOCATE(TEMP_21(zz1,n),TEMP_1(zz1),TEMP_2(zz1),TEMP_3(zz1),TEMP_4(zz2))
  ALLOCATE(TEMP_5(zz2,Ksupepsilon),TEMP_6(zz1,zz1),TEMP_7(zz2,zz2))



  ALLOCATE(sam_thetasubX(saved,zz1),sam_thetasubepsilon(saved,zz1),sam_thetasubb(saved,zz2))
  ALLOCATE(sam_sigma2c(saved),sam_sigma2d(saved))
  ALLOCATE(sam_sigma2epsilon(saved,n),sam_sigma2b(saved,Ksupepsilon))


  ALLOCATE(metro6(n),metro7(Ksupepsilon)) ! don't forget to intialize these to 0 afterwards


  metro6 = 0.; metro7 = 0. 






  WRITE(*,*) "Name of the file containing the X and Y vector (length of name <= 50)"
  READ(*,*) fname1, fname2
  X = readbuff(fname1,n)
  Y = readbuff(fname2,n)

  WRITE(*,*) "Knots files for mean and penalty functions respectively (length of name <= 50)"
  READ(*,*) fname1, fname2
  knotsm = readbuff(fname1,Ksupepsilon)
  knotsb = readbuff(fname2,Ksupb)





!!!!!!!! MAIN BODY OF THE CODE STARTS HERE 


  CsubX(1:n,1) = 1.0
  CsubX(1:n,2) = X
  CsubX(1:n,3) = X**2  

  DO iter1 = 1,n
     DO iter2 = 1,Ksupepsilon
        CsubX(iter1,iter2+p+1) = (DIM(X(iter1),knotsm(iter2)))**2
     end DO
  end DO


  Csubepsilon = CsubX ! Csubepsilon is the same as CsubX

  transCsubX = transpose(CsubX)
  transCsubepsilonbysigma2u = transpose(Csubepsilon)/sigma2u
  CTCepsilonbysigma2u = MATMUL(transCsubepsilonbysigma2u,Csubepsilon)


  Csubb(1:Ksupepsilon,1) = 1.0
  Csubb(1:Ksupepsilon,2) = knotsm

  DO iter1 = 1,Ksupepsilon
     DO iter2 = 1,Ksupb
        Csubb(iter1,iter2+2) = DIM(knotsm(iter1),knotsb(iter2))
     end DO
  end DO
  
  transCsubbbysigma2v = transpose(Csubb)/sigma2v
  CTCbbysigma2v = MATMUL(transCsubbbysigma2v,Csubb)


  gengamvar1 = asubc+Ksupepsilon/2.  
  gengamvar2 = asubd+Ksupb/2. 



  DO iter1 = 1,p+1
     inv_SIGMAsub0X(iter1,iter1) = 1./sigma20beta
     inv_SIGMAsub0epsilon(iter1,iter1) = 1./sigma20gamma
  end DO

  DO iter1 = 1,q+1
     inv_SIGMAsub0b(iter1,iter1) = 1./sigma20delta
  end DO





  CALL time_set_seeds !!! set the seed according to the system clock





!!! Getting Initial Values Now

  WRITE(*,*)""
  WRITE(*,*) "Now we need to get the initial value files. Enter names of initial value files for (length of name < = 50):"
  WRITE(*,*) "\sigma^2_\epsilon, \sigma^2_b, \\theta_X, \\theta_\epsilon, \\theta_b"
  READ(*,*) fname1, fname2, fname3, fname4, fname5
  sigma2epsilon = readbuff(fname1,n)
  sigma2b = readbuff(fname2,Ksupepsilon)
  thetasubX = readbuff(fname3,Ksupepsilon+p+1)
  thetasubepsilon = readbuff(fname4,Ksupepsilon+p+1)
  thetasubb = readbuff(fname5,Ksupb+q+1)

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!                                                                     !!!!!
!!!!!                     STARTING THE MCMC SAMPLER                       !!!!!
!!!!!                                                                     !!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


  WRITE(*,*)""
  WRITE(*,*) "Starting MCMC, Time:", TIME,"s"


  DO main_iter = 1,max_iter  ! main iteration variable for the code; the scheme starts from step 4 of the algorithm

     
!!! STEPS 4 & 5 Generating sigma2c and sigma2d

     reci_sigma2c = bsubc+dot_product(thetasubepsilon(p+2:Ksupepsilon+p+1),thetasubepsilon(p+2:Ksupepsilon+p+1))/2.
     reci_sigma2c = real(random_gamma(real(reci_sigma2c),real(gengamvar1)),kn)

     reci_sigma2d = bsubd+dot_product(thetasubb(q+2:Ksupb+q+1),thetasubb(q+2:Ksupb+q+1))/2.
     reci_sigma2d = real(random_gamma(real(reci_sigma2d),real(gengamvar2)),kn)



     
!!! Calculating matrices mu, neta and si --> needed for steps 6 & 7

     mu = MATMUL(CsubX,thetasubX)
     eta = MATMUL(Csubepsilon,thetasubepsilon) 
     si = MATMUL(Csubb,thetasubb)
     
     
     
     
!!! STEP 6 --> M-H Move to generate sigma2epsilon for each data point
     
     DO iter1 = 1,n
        a = sigma2epsilon(iter1)
        c = eta(iter1)
        d = Y(iter1)
        e = mu(iter1)
        mhprop = real(random_normal(real(a),.04),kn)
           IF (mhprop > 0) THEN  ! proposal can be accepted only if it is positive; otherwise the metropolis ratio would be 0 
              b = (mhprop/a)**(-1.5) * exp((((d-e)**2)/2.)*(1./a-1./mhprop))
              b = b * exp((1./(2.*sigma2u))*(((log(a)-c)**2)-((log(mhprop)-c)**2)))
              IF (b >= 1) THEN
                 a = mhprop  ! accept if metropolis ratio is greater than or equal to 1
                 metro6(iter1)=metro6(iter1)+1
              ELSE
                 u = real(random_standard_uniform(),kn) ! generate and compare to U(0,1) here 
                 IF(u < b) THEN 
                    a = mhprop
                    metro6(iter1)=metro6(iter1)+1
                 ELSE 
                 END IF
              end IF
           ELSE
           end IF
        sigma2epsilon(iter1) = a
     end DO
     
     


!!! STEP 7 --> M-H Move to generate sigma2b for each data point


     DO iter1 = 1,Ksupepsilon
        a = sigma2b(iter1)
        c = si(iter1)
        d = thetasubX(iter1+p+1)
           IF (iter1<=8) THEN
              mhprop = real(random_normal(real(a),525.),kn) 
           ELSE IF ((iter1>8).AND.(iter1<=16)) THEN
              mhprop = real(random_normal(real(a),1300.),kn) 
           ELSE IF ((iter1>16).AND.(iter1<=20)) THEN
              mhprop = real(random_normal(real(a),3000.),kn) 
           ELSE IF ((iter1>20).AND.(iter1<=22)) THEN
              mhprop = real(random_normal(real(a),6000.),kn) 
           ELSE IF ((iter1>22).AND.(iter1<=25)) THEN
              mhprop = real(random_normal(real(a),10000.),kn) 
           ELSE IF ((iter1>25).AND.(iter1<=29)) THEN
              mhprop = real(random_normal(real(a),20000.),kn) 
           ELSE IF ((iter1>29).AND.(iter1<=32)) THEN
              mhprop = real(random_normal(real(a),40000.),kn) 
           ELSE IF ((iter1>32).AND.(iter1<=37)) THEN
              mhprop = real(random_normal(real(a),60000.),kn) 
           ELSE 
              mhprop = real(random_normal(real(a),150000.),kn) 
           end IF
           IF (mhprop > 0) THEN  ! proposal can be accepted only if it is positive; otherwise the metropolis ratio would be 0 
              b = (mhprop/a)**(-1.5)
              b = b * exp((((d**2)/2.)*(1./a-1./mhprop))+((1./sigma2v)*(((log(a)-c)**2)-((log(mhprop)-c)**2))))
              IF (b >= 1) THEN
                 a = mhprop  ! accept if metropolis ratio is greater than or equal to 1
                 metro7(iter1)=metro7(iter1)+1
              ELSE
                 u = real(random_standard_uniform(),kn) ! generate and compare to U(0,1) here 
                 IF(u < b) THEN 
                    a = mhprop 
                    metro7(iter1)=metro7(iter1)+1
                 ELSE
                 end IF
              end IF
           ELSE
           end IF
        sigma2b(iter1) = a
     end DO







!!! Computing inv_SIGMA matrices for STEPS 1, 2 and 3
     
     DO iter1 = 1,Ksupepsilon
        inv_SIGMAsub0X(iter1+p+1,iter1+p+1) = 1./sigma2b(iter1)
        inv_SIGMAsub0epsilon(iter1+p+1,iter1+p+1) = reci_sigma2c
     end DO

     DO iter1 = 1,Ksupb
        inv_SIGMAsub0b(iter1+q+1,iter1+q+1) = reci_sigma2d
     end DO




!!! Calculation of inv_MsubX using the special form of CsubX


     TEMP_21 = multdiag_right(transCsubX,1./sigma2epsilon) ! TEMP_21 stores transpose(CsubX)*inv_SIGMAsubepsilon
     TEMP_2 = MATMUL(transCsubepsilonbysigma2u,log(sigma2epsilon))
     TEMP_3 = MATMUL(TEMP_21,Y)
     inv_MsubX = MATMUL(TEMP_21,CsubX)+inv_SIGMAsub0X



 !!! Computing the inv_Msubepsilon matrix

      inv_Msubepsilon = CTCepsilonbysigma2u + inv_SIGMAsub0epsilon !! intrinsic operation here is well optimized to beat code to sum up just the diagonal elements




 !!! Calculating the mean vector for the multivariate normal density of STEP 1 --> storing it in TEMP_1

      CALL multinvpdsymm_left(inv_MsubX,TEMP_3,TEMP_1,TEMP_6)



!!! STEP 1

      DO iter1 = 1,Ksupepsilon+p+1
         thetasubX(iter1) = random_standard_normal()
      end DO

      thetasubX = TEMP_1 + multinvupper_left(TEMP_6,thetasubX) 



!!! Calculating the mean vector for the multivariate normal density of STEP 2 --> storing it in TEMP_1


      CALL multinvpdsymm_left(inv_Msubepsilon,TEMP_2,TEMP_1,TEMP_6) 


!!! STEP 2

      DO iter1 = 1,Ksupepsilon+p+1
         thetasubepsilon(iter1) = random_standard_normal()
      end DO
      thetasubepsilon = TEMP_1 + multinvupper_left(TEMP_6,thetasubepsilon) 



     inv_Msubb = CTCbbysigma2v + inv_SIGMAsub0b



!!! Calculating the mean vector for the multivariate normal density of STEP 3 --> storing it in TEMP_4

     CALL multinvpdsymm_left(inv_Msubb,transCsubbbysigma2v,TEMP_5,TEMP_7)
     TEMP_4 = MATMUL(TEMP_5,log(sigma2b))


!!! STEP 3

      DO iter1 = 1,Ksupb+q+1
         thetasubb(iter1) = random_standard_normal()
      end DO
      thetasubb = TEMP_4 + multinvupper_left(TEMP_7,thetasubb) 


     IF ((main_iter > burn_in).AND.(MOD(main_iter,thin) == 0)) THEN ! save the data when these conditions are met
        save_iter = save_iter + 1

        sam_thetasubX(save_iter,1:Ksupepsilon+p+1) = thetasubX ! saving thetasubX
        sam_thetasubepsilon(save_iter,1:Ksupepsilon+p+1) = thetasubepsilon ! saving thetasubepsilon
        sam_thetasubb(save_iter,1:Ksupb+q+1) = thetasubb ! saving thetasubb
        sam_sigma2c(save_iter) = 1./reci_sigma2c ! saving sigma2c
        sam_sigma2d(save_iter) = 1./reci_sigma2d ! saving sigma2d
        sam_sigma2epsilon(save_iter,1:n) = sigma2epsilon 
        sam_sigma2b(save_iter,1:Ksupepsilon) = sigma2b 
     end IF

     IF (MOD(main_iter,5000) == 0) THEN
        CALL CPU_TIME(TIME)
        WRITE(*,*) "Iteration:", main_iter, ", Time:", TIME,"s"
     end IF

!     WRITE(*,*) main_iter

  end DO  ! closing the main iteration loop 


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!                                                                     !!!!!
!!!!!      END OF SAMPLER. DEALLOCATING MATRICES AND WRITING OUTPUT       !!!!!
!!!!!                                                                     !!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  DEALLOCATE(thetasubX,thetasubepsilon,thetasubb,Y,sigma2epsilon,mu,eta,X,knotsm,knotsb)
  DEALLOCATE(sigma2b,si,Csubepsilon,transCsubX,transCsubepsilonbysigma2u,Csubb) 
  DEALLOCATE(transCsubbbysigma2v,inv_MsubX,inv_Msubepsilon,CTCepsilonbysigma2u,inv_SIGMAsub0X,inv_SIGMAsub0epsilon,inv_SIGMAsub0b)
  DEALLOCATE(inv_Msubb,CTCbbysigma2v,TEMP_21,TEMP_1,TEMP_2,TEMP_3,TEMP_4,TEMP_5,TEMP_6)
  
  CALL writebuff("sam_m",MATMUL(CsubX,transpose(sam_thetasubX)))
  CALL writebuff("sam_thetasubX",sam_thetasubX)
  DEALLOCATE(sam_thetasubX,CsubX)

  CALL writebuff("sam_thetasubepsilon",sam_thetasubepsilon)
  DEALLOCATE(sam_thetasubepsilon)

  CALL writebuff("sam_thetasubb",sam_thetasubb)
  DEALLOCATE(sam_thetasubb)

  CALL writebuff("sam_sigma2c",sam_sigma2c)
  DEALLOCATE(sam_sigma2c)

  CALL writebuff("sam_sigma2d",sam_sigma2d)
  DEALLOCATE(sam_sigma2d)

  CALL writebuff("sam_sigma2b",sam_sigma2b)
  DEALLOCATE(sam_sigma2b)

  CALL writebuff("sam_sigma2epsilon",sam_sigma2epsilon)
  DEALLOCATE(sam_sigma2epsilon)

  CALL writebuff("metro6",metro6/max_iter)
  DEALLOCATE(metro6)

  CALL writebuff("metro7",metro7/max_iter)
  DEALLOCATE(metro7)

end PROGRAM bayes_pspline_hetero
