
MODULE matrix
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!                                                                         !!!
!!!                                                                         !!!
!!! THIS MODULE HAS BEEN DESIGNED SPECIFICALLY TO ACHIVE FASTER COMPUTATION !!!
!!! OF CERTAIN MATRIX OPERATIONS. DETAILS OF WHAT EACH ROUTINE DOES ARE     !!!
!!! GIVEN ALONGWITH ITS DECLARATION.                                        !!!
!!!                                                                         !!!
!!!                                                                         !!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  INTERFACE multinvupper_left
     MODULE PROCEDURE multinvupper_left_v,multinvupper_left_m
  end INTERFACE

  INTERFACE multinvpdsymm_left
     MODULE PROCEDURE multinvpdsymm_left_mo, multinvpdsymm_left_vo
  end INTERFACE

  INTERFACE printmatrix
     MODULE PROCEDURE printmatrix_m,printmatrix_v
  end INTERFACE


CONTAINS

!!! ALWAYS INITIALIZE MATRICES IN SUBROUTINES


  function identity(rank) result(A)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!                                                                         !!!
!!!            RETURNS IDENTITY MATRIX OF DIMENSION "rank"                  !!!
!!!                                                                         !!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    use kind
    IMPLICIT NONE
    INTEGER :: rank, i, j
    REAL(kn),DIMENSION(rank,rank) :: A
    DO i=1,rank
       DO j=1,rank
          A(i,j) = 0
       end DO
       A(i,i) = 1
    end DO
  end function identity



  SUBROUTINE printmatrix_m(A)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!                                                                         !!!
!!!                     PRINTS GIVEN MATRIX AS OUTPUT                       !!!
!!!                                                                         !!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    use kind
    IMPLICIT NONE
    REAL(kn),DIMENSION(:,:) :: A
    INTEGER :: i,j
    DO i = 1,size(A,1)
       WRITE(*,*) (A(i,j),j=1,size(A,2))
    end DO
    WRITE(*,*) ""
  end SUBROUTINE printmatrix_m

  SUBROUTINE printmatrix_v(A)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!                                                                         !!!
!!!                     PRINTS GIVEN VECTOR AS OUTPUT                       !!!
!!!                                                                         !!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    use kind
    IMPLICIT NONE
    REAL(kn),DIMENSION(:) :: A
    INTEGER :: i
    DO i = 1,size(A,1)
       WRITE(*,*) A(i)
    end DO
    WRITE(*,*) ""
  end SUBROUTINE printmatrix_v



  function choldc(a) result(b) 
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!                                                                         !!!
!!!     RETURNS UPPER TRIANGULAR CHOLESKY FACTOR OF GIVEN MATRIX A          !!!
!!!                                                                         !!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    use kind
    IMPLICIT NONE
    REAL(kn), DIMENSION(:,:), INTENT(IN) :: a
    REAL(kn), DIMENSION(size(a,1),size(a,1)) :: b
    INTEGER :: i,j,n
    b = 0
    n = size(a,1)
    b(1,1) = sqrt(a(1,1))
    b(1,2:n) = a(1,2:n)/b(1,1)
    do i=2,n
       b(i,i:n) = a(i,i:n)-matmul(b(1:i-1,i),b(1:i-1,i:n))
       if (b(i,i) <= 0.0) THEN
          WRITE(*,*) "Error: Matrix in cholesky decomposition is not symmetric, positive definite" ! a, with rounding errors, is not symmetric positive definite 
          STOP
       end if
       b(i,i) = sqrt(b(i,i))
       b(i,i+1:n) = b(i,i+1:n)/b(i,i)
    end do
  end function choldc



  function cholsl(a,b) result(x)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!                                                                         !!!
!!!   RETURNS BY BACK-SUBSTITUTION THE SOLUTION x SYSTEM OF EQUATIONS:      !!!
!!!                       ax =  b                                           !!!
!!!   HERE a IS UPPER TRIANGULAR AND b IS A VECTOR.                         !!!
!!!                                                                         !!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    use kind
    IMPLICIT NONE
    REAL(kn), DIMENSION(:,:), INTENT(IN) :: a
    REAL(kn), DIMENSION(:), INTENT(IN) :: b
    REAL(kn), DIMENSION(size(b,1)) :: x
    INTEGER :: i,n
    n = size(a,1)
    do i=1,n
       x(i)=(b(i)-dot_product(a(1:i-1,i),x(1:i-1)))/a(i,i)
    end do
    do i=n,1,-1
       x(i)=(x(i)-dot_product(a(i,i+1:n),x(i+1:n)))/a(i,i)
    end do
  end function cholsl


  subroutine multinvpdsymm_left_mo(A,B,C,U) 
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!                                                                         !!!
!!!   GIVEN MATRICES A AND B,WHERE A IS POSITIVE DEFINITE AND SYMMETRIC,    !!!
!!!   THIS SUBROUTINE COMPUTES MATRICES U AND C. U IS THE UPPER             !!!
!!!   TRIANGULAR CHOLESKY FACTOR OF A AND C:                                !!! 
!!!                       C =  inv(A)B                                      !!!
!!!                                                                         !!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    use kind
    IMPLICIT NONE
    INTEGER :: j,k,n
    REAL(kn),DIMENSION(:,:),INTENT(IN) :: A
    REAL(kn),DIMENSION(:,:),INTENT(IN) :: B
    REAL(kn),DIMENSION(size(A,1),size(A,1)) :: U
    REAL(kn),DIMENSION(size(B,1),size(B,2)) :: C
    n =size(C,2)
    U = real(choldc(A))
    DO j = 1,n
       C(1:size(C,1),j) = cholsl(U,B(1:size(C,1),j))
    end DO
  end subroutine multinvpdsymm_left_mo


  subroutine multinvpdsymm_left_vo(A,B,C,U)  
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!                                                                         !!!
!!!      SAME AS multinvpdsymm_left_mo EXCEPT THAT B IS A VECTOR NOW        !!!   
!!!                                                                         !!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    use kind
    IMPLICIT NONE
    REAL(kn),DIMENSION(:,:),INTENT(IN) :: A
    REAL(kn),DIMENSION(:), INTENT(IN) :: B
    REAL(kn),DIMENSION(size(A,1),size(A,1)) :: U
    REAL(kn),DIMENSION(size(B,1)) :: C 
    U = real(choldc(A))
    C = cholsl(U,B)
  end subroutine multinvpdsymm_left_vo



  function multdiag_left(A,B) result(C) 
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!                                                                         !!!
!!! TAKES MATRIX B AND PREMULTIPLIES IT WITH A DIAGONAL MATRIX WHOSE        !!!
!!! DIAGONAL VECTOR IS A. B NEED NOT BE A SQUARE MATRIX                     !!!     
!!!                                                                         !!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    use kind
    IMPLICIT NONE
    REAL(kn), DIMENSION(:) :: A
    REAL(kn), DIMENSION(:,:) :: B
    REAL(kn),DIMENSION(size(B,1),size(B,2)) :: C
    INTEGER :: i
    IF (size(B,1) /= size(A,1)) THEN
       WRITE(*,*) "Error in the function multdiag_left: Dimensions of the diagonal vector don't match the dimensions of the matrix"
       STOP
    end IF
    DO i = 1,size(B,1)
       C(i,1:size(B,2)) = A(i)*B(i,1:size(B,2))
    end DO
  end function multdiag_left


  function multdiag_right(B,A) result(C)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!                                                                         !!!
!!! TAKES MATRIX B AND POSTMULTIPLIES IT WITH A DIAGONAL MATRIX WHOSE       !!!
!!! DIAGONAL VECTOR IS A. B NEED NOT BE A SQUARE MATRIX                     !!!     
!!!                                                                         !!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    use kind
    IMPLICIT NONE
    REAL(kn), DIMENSION(:) :: A
    REAL(kn), DIMENSION(:,:) :: B
    REAL(kn),DIMENSION(size(B,1),size(B,2)) :: C
    INTEGER :: i
    IF (size(B,2) /= size(A,1)) THEN
       WRITE(*,*) "Error in the function multdiag_right: Dimensions of the diagonal vector don't match the dimensions of the matrix"
       STOP
    end IF
    DO i = 1,size(B,2)
       C(1:size(B,1),i) = A(i)*B(1:size(B,1),i)
    end DO
  end function multdiag_right
  


  function multinvupper_left_v(U,A) result(B) 
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!                                                                         !!!
!!! FOR UPPER TRIANGULAR MATRIX U AND A VECTOR A, THIS FUNCTION RETURNS B:  !!!
!!!                             B = inv(U)A                                 !!!     
!!!                                                                         !!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    use kind
    IMPLICIT NONE
    REAL(kn), DIMENSION(:,:) :: U
    REAL(kn), DIMENSION(:) :: A
    REAL(kn), DIMENSION(size(A,1)) :: B
    INTEGER :: i,n
    n = size(A,1)
    DO i = n,1,-1
       IF(U(i,i)==0) pause 'Singular U in multinvipper_left_v'
       B(i) = (A(i)-dot_product(U(i,i+1:n),B(i+1:n)))/U(i,i)
    end DO
  end function multinvupper_left_v
  

  function multinvupper_left_m(U,A) result(B) 
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!                                                                         !!!
!!! SAME AS multinvupper_left_v EXCEPT THAT B IS A MATRIX NOW               !!!
!!!                                                                         !!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    use kind
    IMPLICIT NONE
    REAL(kn), DIMENSION(:,:) :: U
    REAL(kn), DIMENSION(:,:) :: A
    REAL(kn), DIMENSION(size(A,1),size(A,2)) :: B
    INTEGER :: i,j,n,m
    n = size(A,1)
    m = size(A,2)
    DO j = 1,m
       DO i = n,1,-1
          IF(U(i,i)==0) pause 'Singular U in multinvipper_left_m'
          B(i,j) = (A(i,j)-dot_product(U(i,i+1:n),B(i+1:n,j)))/U(i,i)
       end DO
    end DO
  end function multinvupper_left_m


end MODULE matrix
